/***************************************************************************
                          ringbuffer.h  -  description
                             -------------------
    begin                : Tue Apr 17 2001
    copyright            : (C) 2001 by Martin Schulze
    email                : martin-ml@hippogriff.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __ringbuffer_h
#define __ringbuffer_h


#ifdef DEBUG
#include <iostream>
#endif
#include <utility>
#include <limits.h>
#include "bitops.h" //<asm/bitops.h>



namespace Util {


/***************************************************************************
 * ringbuffer_: base class for ringbuffer
 ***************************************************************************/
template <class T,bool dynamic>
struct destructor {
  void destroy( T obj )   {}
};

template <class T>
struct destructor<T,true> {
  void destroy( T obj )   { delete obj; }
};

struct ringbuffer_ {
  typedef unsigned char BYTE;

  ringbuffer_( BYTE* vchilds, unsigned short vchldSize, unsigned short vrealSize );
                      // isOwner: when true the elements get deleted automatically
                      // aSize: (fixed) maximal count of elements
  ~ringbuffer_();

  unsigned short push();
  void push_finish( unsigned short pos );
  unsigned short pop();
  void pop_finish( unsigned short pos );    // gibt eine Position für NewMember frei

  BYTE* at( unsigned short pos );
  const BYTE* at( unsigned short pos ) const;

  // Positionen
  unsigned short first() const;
  unsigned short last() const;
  unsigned short incr( unsigned short pos, unsigned short count ) const;  // count Richtung LastPos
  unsigned short decr( unsigned short pos, unsigned short count ) const; // count Richtung FirstPos
  unsigned short incr( unsigned short& pos ) const;  // count Richtung LastPos
  unsigned short decr( unsigned short& pos ) const; // count Richtung FirstPos

  // Zerstoerung
  void clear();

  // Info
  unsigned short size() const               { return( size_ <= 0 ? 0 : (unsigned short)size_ ); }
  bool empty() const                        { return( size_ <= 0 ); }
  bool full() const                         { return( size_ == realSize-1 ); }
  bool valid_position( int vpos ) const;
  bool valid( int pos ) const;

// Variablen:
private:
  BYTE* childs;
  BYTE* appendMask;
  BYTE* removeMask;
  volatile int firstPos;
  volatile int lastPos;
  volatile int size_;

public:
  const unsigned short chldSize;
  const unsigned short realSize;
};


/***************************************************************************
 * ringbuffer: container class with fifo-logic and thread-safe push/pop
 ***************************************************************************/
template <class type,unsigned short _size_=50,bool dynamic=false>
class ringbuffer {
  ringbuffer_ _rb_;

public:
  ringbuffer();
  ~ringbuffer();

// Funktionen:
  // Hinzufuegen / Entfernen
  void clear();
  bool push( const type& member );
  std::pair<type,bool> pop();
  destructor<type,dynamic> garbage;

  // Info
  unsigned short size() const               { return _rb_.size(); }
  bool empty() const                        { return _rb_.empty(); }
  bool full() const                         { return _rb_.full(); }
};


inline ringbuffer_::BYTE* ringbuffer_::at( unsigned short pos )
  { return( &childs[ pos * chldSize ] ); }

inline const ringbuffer_::BYTE* ringbuffer_::at( unsigned short pos ) const
  { return( &childs[ pos * chldSize ] ); }

inline unsigned short ringbuffer_::first() const {
  int temp = firstPos;
  while( temp >= realSize ) temp -= realSize;
  while( temp < 0 ) temp += realSize;
  return temp;
}

inline unsigned short ringbuffer_::last() const {
  int temp = lastPos;
  while( temp >= realSize ) temp -= realSize;
  while( temp < 0 ) temp += realSize;
  return temp;
}

inline unsigned short ringbuffer_::incr( unsigned short pos, unsigned short count ) const {
  if( ( pos >= realSize ) || ( count > realSize ) ) return USHRT_MAX;
  return( pos >= (unsigned short)(realSize-count) ? count - (realSize-pos) : pos + count );
}

inline unsigned short ringbuffer_::decr( unsigned short pos, unsigned short count ) const {
  if( count > realSize ) return USHRT_MAX;
  return( pos < count ? realSize - (count-pos) : pos - count );
}

inline unsigned short ringbuffer_::incr( unsigned short& pos ) const {
  if( ++pos >= realSize ) pos -= realSize;
  return pos;
}

inline unsigned short ringbuffer_::decr( unsigned short& pos ) const {
  if( pos == 0 ) pos = realSize;
  return( --pos );
}

inline bool ringbuffer_::valid_position( int pos ) const {
//  if( size_ == realSize ) return( pos < size_ );  // there is always an unused placeholder => firstPos != lastPos
  if( lastPos >= firstPos ) return( ( pos >= firstPos ) && ( pos < lastPos ) );
  else return( ( pos >= firstPos ) || ( pos < lastPos ) );
}

inline bool ringbuffer_::valid( int pos ) const {
  if( !valid_position( pos ) ) return false;
  return( test_bit( pos, removeMask ) ); //  test_bit( readMask ) && !test_bit( removeMask ) ????
}


template <class type,unsigned short _size_,bool dynamic>
inline ringbuffer<type,_size_,dynamic>::ringbuffer() :
  _rb_( (ringbuffer_::BYTE*)( new type[_size_+1] ), sizeof( type ), _size_+1 )  {}

template <class type,unsigned short _size_,bool dynamic>
inline ringbuffer<type,_size_,dynamic>::~ringbuffer()
  { clear(); delete [] (type*)( _rb_.at(0) ); }

template <class type,unsigned short _size_,bool dynamic>
inline void ringbuffer<type,_size_,dynamic>::clear() {
  if( dynamic ) {
    for( unsigned short i = 0; i < _size_; ++i )
      if( _rb_.valid(i) ) garbage.destroy( *reinterpret_cast<type*>( _rb_.at( i ) ) );
  }
  _rb_.clear();
}

template <class type,unsigned short _size_,bool dynamic>
inline bool ringbuffer<type,_size_,dynamic>::push( const type& member ) {
  unsigned short pos = _rb_.push();
  if( pos == USHRT_MAX ) {
    if( dynamic ) garbage.destroy( member );
    return false;
  } else {
    *reinterpret_cast<type*>( _rb_.at( pos ) ) = member;
    _rb_.push_finish( pos );
    return true;
  }
}

template <class type,unsigned short _size_,bool dynamic>
inline std::pair<type,bool> ringbuffer<type,_size_,dynamic>::pop() {
  unsigned short pos = _rb_.pop();
  if( pos == USHRT_MAX ) return std::pair<type,bool>( type(), false );
  type temp = *reinterpret_cast<type*>( _rb_.at( pos ) );
  _rb_.pop_finish( pos );
  return std::pair<type,bool>( temp, true );
}

}


#endif
